#!/usr/bin/env python


import math
import random
import q_classifier_r
import g_implements
import Num
import gpkmisc
import Numeric_gpk as NG
import fiatio
import chunkio
import die

HUGE = 1e30
Clip = None

def find_which_cluster(datum, cpos, distfcn):
	bestdist = HUGE
	best = -1
	for (i, cp) in enumerate(cpos):
		tmp = distfcn(datum.value, cp)
		if tmp < bestdist:
			bestdist = tmp
			best = i
	return (best, bestdist)


def find_center(i, membership, data, clip):
	tmp = []
	for (mem, datum) in enumerate(membership, data):
		if mem == i:
			tmp.append( datum.value )
	ctr = NG.trimmed_mean_across(tmp, None, clip)
	return ctr


class cluster_descriptor:
	def __init__(self, center, membership, size):
		self.center = center
		self.membership = membership
		self.size = size



def kmeans(data, Ncl, distfcn=None, Nit=None, clip=None):
	for datum in data:
		g_implements.check(datum, q_classifier_r.datum_c)
	
	if Nit is None:
		Nit = 10 + int(round(math.sqrt(len(data)*Ncl)))
	if clip is None:
		clip = 0.10
	assert len(data) >= Ncl
	bestcp = None
	bestm = None
	besterr = HUGE
	starts = 0
	while starts < Nit:
		cpos = [data[c].value for c in random.sample(range(len(data)), Ncl) ]
		membership = Num.zeros((len(data),), Num.Int)
		assert len(cpos) == Ncl
		passes = 0
		while passes < Nit:
			cldtmp = [ [] for x in range(Ncl) ]
			scatter = 0.0
			omem = Num.array(membership, copy=True)
			for (i, datum) in enumerate(data):
				membership[i], dist = find_which_cluster(datum, cpos, distfcn)
				if membership[i] >= 0:
					cldtmp[membership[i]].append(dist)
				scatter += dist
			cldist = [ gpkmisc.median(cldtmp[i]) for i in range(Ncl) ]

			for i in range(Ncl):
				cpos[i] = find_center(i, membership, data, clip)
			if omem == membership:
				break
			passes += 1

		if scatter < besterr:
			besterr = scatter
			bestcp = cpos
			bestm = membership

		starts += 1

	o = []
	for i in range(Ncl):
		m = [ d.uid for (mem, d) in zip(bestm, data) if mem == i ]
		o.append( cluster_descriptor(bestcp[i], m, cldist[i]) )
	map_to_cluster = {}
	for (datum, mem) in zip(data, bestm):
		map_to_cluster[datum] = mem
	return (o, map_to_cluster, besterr)




def read_data(fd):
	"""Reads in feature vectors, each with a uid as a comment."""
	d = []
	dim = None
	# die.info('Reading')
	ln = 0
	# for l in fd.readlines():
	for l in fd:
		ln += 1
		if l.startswith('#'):
			continue
		aa = l.split('#', 1)
		if len(aa) > 1:
			uid = aa[1].strip()
		else:
			uid = 'Line:%d' % ln
		a = aa[0].split()
		if dim is None:
			dim = len(a)
		elif len(a) != dim:
			die.die('Not all vectors have length=%d.  Problem on line %d'
					% (dim, ln)
				)
		d.append( q_classifier_r.datum_c(Num.array([float(x) for x in a]),
				uid)
				)
	return d


def euclid(a, b):
	return math.sqrt(Num.sum(Num.ravel(a-b)**2))


if __name__ == '__main__':
	import sys
	arglist = sys.argv[1:]
	hdrs = {}
	uidname = 'uid'
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-header':
			k = arglist.pop(0)
			v = arglist.pop(0)
			hdrs[k] = v
		elif arg == '-uidname':
			uidname = arglist.pop(0)
		elif arg == '-clip':
			Clip = float(arglist.pop(0))
		else:
			die.die('Unrecognized flag: %s' % arg)
	ncl = int(arglist[0])
	d = read_data(sys.stdin)
	o, map_to_cluster, err = kmeans(d, ncl, distfcn=euclid, clip=Clip)
	w = fiatio.writer(sys.stdout)
	w.headers(hdrs)
	w.header('Ncl', ncl)
	if Clip is not None:
		w.header('Clip', Clip)
	w.header('Error', err)
	for (clnum, descr) in enumerate(o):
		w.header('cluster%d' % clnum,
			chunkio.chunkstring_w().write_NumArray(descr.center, str).close()
			)
		w.header('size%d' % clnum, descr.size)
	for (clnum, descr) in enumerate(o):
		for datum_uid in descr.membership:
			tmp = {'clnum': clnum, uidname: datum_uid}
			w.datum(tmp)
	w.close()



"""This is a helper module to make use of mcmc.py and mcmc_big.py.
It allows you to conveniently run a Monte-Carlo simulation of any
kind until it converges (L{stepper.run_to_bottom}) or until it
has explored a large chunk of parameter space (L{stepper.run_to_ergodic}).

It also helps you with logging the process.
"""

import thread as Thr
import Num
import g_implements
import mcmc

NW = 20
NWE = 300
ERGL = 0.10
NSF = 20
EWC = 100


class TooManyLoops(Exception):
	def __init__(self, *s):
		Exception.__init__(self, *s)




class warnevery(object):
	def __init__(self, interval):
		self.c = 0
		self.interval = interval
		self.lwc = 0
	
	def inc(self, incr=1):
		self.c += incr
		if self.c > self.lwc + self.interval:
			self.lwc = self.c
			return 1
		return 0

	def count(self):
		return self.c


class _counter(object):
	def __init__(self):
		self.lock = Thr.allocate_lock()
		self.count = 0

	def update(self, accepted):
		self.lock.acquire()
		self.count += accepted
		self.lock.release()

	def get(self):
		self.lock.acquire()
		tmp = self.count
		self.lock.release()
		return tmp



class _accum_erg(object):
	def __init__(self):
		self.lock = Thr.allocate_lock()
		self.erg = 0.0
		self.last_resetid = -1

	def update(self, x):
		self.lock.acquire()
		tmp = x.reset_id()
		if tmp != self.last_resetid:
			self.erg = 0.0
			self.last_resetid = tmp
		else:
			self.erg += x.ergodic()
		self.lock.release()


	def get(self):
		self.lock.acquire()
		tmp = self.erg
		self.lock.release()
		return tmp


class logger_template(object):
	def add(self, stepperInstance, iter):
		pass

	def close(self):
		pass


class stepper(object):
	def __init__(self, x, maxloops=-1, logger=None):
		self.logger = None	# This sets things up for close()
		self.iter = 0
		self.maxloops = maxloops
		g_implements.check(x, mcmc.stepper)
		self.x = x
		if logger is None:
			self.logger = logger_template()
		else:
			g_implements.check(logger, logger_template)
			self.logger = logger


	def reset_loops(self, maxloops=-1):
		self.maxloops = maxloops


	def run_to_change(self, ncw=1, update_T=False):
		n = 0
		while n < ncw:
			if update_T:
				pre_logp = self.x.current().logp_nocompute()
			else:
				self.x.T = 1.0
			acc = self.x.step()
			if acc and update_T:
				logp = self.x.current().logp_nocompute()
				self.x.T = max(1.0, self.x.T*0.90 + 0.4*(logp-pre_logp))
			n += acc
			self.logger.add(self.x, self.iter)
			self.maxloops -= 1
			if self.maxloops == 0:
				raise TooManyLoops, "run_to_change"
			self.iter += 1
		return self.x.current()


	def close(self):
		"""After calling close(), it is no longer legal
		to call run_to_change(), run_to_ergodic(), or
		run_to_bottom().    Logging is shut down, but the
		contents of the stepper are still available for
		inspection.
		"""
		if self.logger is not None:
			self.logger.close()
			self.logger = None


	def __del__(self):
		self.close()


	def run_to_ergodic(self, ncw=1):
		"""Run the stepper until it has explored all of parameter space
		C{ncw} times (as best as we can estimate).   Note that this
		is a pretty careful routine.   If the stepper finds a better
		region of parameter space so that log(P) improves part way
		through the process, the routine will reset itself and begin
		again.

		@param ncw: how many times (or what fraction) of an ergodic
			exploration to make.
		@type ncw: int.
		@return: a L{position<mcmc.position_base>} at the end of the process.
		"""
		nc = _accum_erg()
		while nc.get() < ncw:
			self.x.step()
			self.logger.add(self.x, self.iter)
			nc.update(self.x)
			self.iter += 1
			self.maxloops -= 1
			if self.maxloops == 0:
				raise TooManyLoops, ("run_to_ergodic", nc.get(), ncw)
		return self.x.current()


	NC = 3
	def run_to_bottom(self, ns=3):
		m = self.x.np
		self.run_to_change(5+2*m, update_T=True)	# get it warmed up.
		before = self.run_to_change(ns, update_T=True)
		mid = self.run_to_change(ns, update_T=True)
		es = 0.0
		lock = Thr.allocate_lock()
		xchanged = Num.zeros(m, Num.Int)
		dotchanged = 0
		last_resetid = self.x.reset_id()
		while Num.sometrue(Num.less(xchanged,self.NC)) or es<1.0 or dotchanged<2*m:
			early = Num.sum(xchanged)<self.NC*m or es<0.8 or dotchanged<m
			try:
				after = self.run_to_change(ns, update_T=early)
			except TooManyLoops, x:
				raise TooManyLoops, ('run_to_bottom: s' % x, self.iter, es, dotchanged, m,
							Num.sum(Num.less(xchanged, self.NC)),
							self.x.current().prms()
							)
			# We keep track of direction reversals to see when
			# we reach the bottom.
			Num.add(xchanged,
				Num.less( (after.vec()-mid.vec())
						* (mid.vec()-before.vec()), 0.0),
				xchanged)

			dotchanged += Num.dot(after.vec()-mid.vec(),
						mid.vec()-before.vec()) < 0.0

			es += self.x.ergodic() * ns

			# print 'LA'
			lock.acquire()
			if self.x.reset_id() != last_resetid:
				# A new maximum: we are probably
				# not stable.  Likely, we are still in the
				# improvement phase.
				last_resetid = self.x.reset_id()
				lock.release()
				es = 0.0
				xchanged[:] = 0
				dotchanged = 0
			else:
				lock.release()

			before = mid
			mid = after
		return self.iter



def test():
	def test_logp(x, c):
		# print '#', x[0], x[1]
		return -(x[0]-x[1]**2)**2 + 0.001*x[1]**2
	x = mcmc.bootstepper(test_logp, Num.array([0.0,2.0]),
				Num.array([[1.0,0],[0,2.0]]))
	thr = stepper(x)
	# nsteps = thr.run_to_bottom(x)
	# print '#nsteps', nsteps
	# assert nsteps < 100
	for i in range(2):
		print 'RTC'
		thr.run_to_change(2)
		print 'RTE'
		thr.run_to_ergodic(1.0)
		print 'DONE'
	thr.close()




if __name__ == '__main__':
	test()



"""This is a support module, used by many types of classifiers."""

import Num
import types
import chunkio
import blue_data_selector
import math
import gpkavg
import die
import mcmc
import mcmc_helper
import g_implements
import fiatio
import random
import dictops
import gpkmisc

ERGCOVER = 4.0
D = 0
TGT = 0.01

class datum_c(object):
	__doc__ = """This is an unclassified datum, either in the
		test or training set."""

	__slots__ = ['value', 'uid']

	def __init__(self, vector, uid=None):
		"""Vector is the feature vector of this datum,
		and uid is an arbitrary identifier."""
		self.value = Num.asarray(vector, Num.Float)
		self.uid = uid

	def __str__(self):
		return "<datum_c fv=%s uid=%s>" % (str(self.value), self.uid)
	__repr__ = __str__


class datum_tr(datum_c):
	__doc__ = """This is a classified datum, presumably in
		the training set."""

	__slots__ = ['value', 'uid', 'classid']

	def __init__(self, vector, classid, uid=None):
		"""Vector is the feature vector of this datum,
		classid is the true class,
		and uid is an arbitrary identifier."""
		datum_c.__init__(self, vector, uid)
		self.classid = classid
	
	def __str__(self):
		return "<datum_tr class=%s fv=%s uid=%s>" % (self.classid,
								str(self.value),
								self.uid)
	__repr__ = __str__


class bluedata:
	__doc__ = """This is an extension of blue_data_selector.bluedata which
		is aware of the classids of data.  It tries to split the
		data evenly for each class."""

	def __init__(self, data, blueness = 2.0):
		self.ntot = 0
		per_class_data = dictops.dict_of_lists()
		for datum in data:
			per_class_data.add(datum.classid, datum)

		self.selector = {}
		for (cid, cldata) in per_class_data.items():
			self.ntot += len(cldata)
			self.selector[cid] = blue_data_selector.bluedata(cldata, blueness)
	

	def _chooseint(self, n):
		"""This chooses integers math.floor(n) or math.ceil(n) with
		appropriate probabilities so that the expectation value of
		_chooseint(n) is n.
		"""
		f = math.floor(n)
		p = n - f
		assert 0.0 <= p <= 1.0
		i = int(f)
		if random.random() < p:
			i += 1
		return i


	def split(self, n):
		frac = float(n)/float(self.ntot)
		npc = {}
		# First, we choose how many items to select from each class.
		while 1:
			# We pick random numbers for each class, then check to see if the
			# total is right.
			ntot = 0
			for (cid, selector) in self.selector.items():
				tmp = self._chooseint(frac*len(selector))
				assert tmp >= 0 and tmp <= len(selector)
				ntot += tmp
				npc[cid] = tmp
			if n-0.5 <= ntot <= n+0.5:
				break

		# Then, we use a blue data selector for each class to pick the right number
		# of items from that class.
		o1 = []
		o2 = []
		for (cid, selector) in self.selector.items():
			t1, t2 = selector.split(npc[cid])
			o1.extend( t1 )
			o2.extend( t2 )
		assert len(o1) == ntot
		# Finally, we shuffle the classes back together.
		random.shuffle(o1)
		random.shuffle(o2)
		# ...and return.
		return (o1, o2)

	
	def __len__(self):
		return self.ntot



def _r1if(delta):
	return 0.5*(1.0 - delta/math.hypot(delta, 1.0))




def prior(training, testing):
	"""This assumes that you choose class C
	with probability 1 if P(C) is the
	biggest among all the classes.
	It is the best strategy, assuming you can't
	see the feature vector.
	"""
	dtr = dictops.dict_of_accums()
	for datum in training:
		dtr.add(datum.classid, 1)

	counts_best = 0.0
	kbest = None
	for (k, counts) in dtr.items():
		if counts > counts_best:
			counts_best = counts
			kbest = k

	chance = 0
	for datum in testing:
		if datum.classid == kbest:
			chance += 1
	return chance/float(len(testing))



def max_correct(training, testing):
	"""This is a hard, conservative upper limit
	for the probability of correct classification.

	One cannot hope to classify an item correctly,
	if the item's class is not represented
	in the training set.

	If the class is present in the training
	set but not in the test set, then what happens?
	It could be buried among items from other classes,
	in which case it wouldn't lead to any false
	predictions.  Consequently, absences from the
	test set doesn't hurt the maximum possible
	classification probability.
	"""
	dtr = dictops.dict_of_accums()
	for datum in training:
		dtr.add(datum.classid, 1)

	impossible = 0
	for datum in testing:
		if not datum.classid in dtr:
			impossible += 1

	return float(len(testing)-impossible)/float(len(testing))






def _addwrongD(wrong, failures):
	for (k, v) in failures.iteritems():
		wrong.add(k, v)


def _addwrong(wrong, failures):
	for t in failures:
		wrong.add(t, 1)


def _doclass(qc, testing, verbose=True):
	"""testing is a list of data.
	qc is a classifier.
	This classifies the data and yields a list of dictionaries,
	each of which contains a bunch of results and information.
	"""
	info = qc.info.copy()
	for datum in testing:
		P = qc.P(datum)
		tmp = info.copy()

		# It is tempting to use qc.bestc() here, but
		# not a good idea, as it would involve calculating
		# log(P) twice, and that can be expensive.
		bestcl = None
		bestp = 0.0
		for (cl, p) in P:
			if p > bestp:
				bestp = p
				bestcl = cl
		tmp['compclass'] = bestcl

		if datum.uid is not None:
			tmp['Duid'] = datum.uid

		if verbose:
			tmp['V'] = chunkio.chunkstring_w().write_NumArray(datum.value, str).close()
			tmp['P'] = chunkio.chunkstring_w().write_dict(dict(P)).close()
		try:
			tmp['trueclass'] = datum.classid
		except AttributeError:
			pass
		yield tmp




def read_data(fd, commentarray=None):
	"""Reads in feature vectors where the first element is the true class."""
	d = []
	len_a = None
	# die.info('Reading')
	print '# Reading'
	ln = 0
	for l in fd:
		ln += 1
		if l.startswith('#'):
			if commentarray is not None:
				commentarray.append(l[1:])
			continue
		aa = l.split('#', 1)
		if len(aa) > 1:
			uid = aa[1].strip()
		else:
			uid = 'Line:%d' % ln
		a = aa[0].split()
		if len(a) != len_a:
			if len_a is None:
				len_a = len(a)
			else:
				die.die('Not all vectors have length=%d.  Problem on line %d'
						% (len_a-1, ln)
					)
		d.append( datum_tr(Num.array([float(x) for x in a[1:]]),
				a[0], uid)
				)
	return d


def get_dim(fd):
	"""This function takes a list of data (type datum_tr)
	and makes sure that they all have the same length feature
	vector.  If so, it reports the length (dimension) of the
	feature vector.
	"""
	assert type(fd) == types.ListType
	dim = None
	for x in fd:
		assert isinstance(x, datum_tr)
		if len(x.value) != dim:
			if dim is None:
				dim = len(x.value)
			assert len(x.value) == dim
	return dim
		


def compute_cross_class(training, testing,
			modelchoice=None, n_per_dim=None,
			builder=None, classout=None,
			trainingset_name = None,
			verbose=True):
	"""Build classifiers based on the training set,
	and test them on the testing set.
	Modelchoice here is the completed class object,
	not a closure."""

	if len(training) == 0:
		die.die('No training data.')
	if len(testing) == 0:
		die.die('No data to classify.')

	dim = get_dim(training)
	assert get_dim(testing) == dim
	if dim <= 0:
		die.die('zero dimensional data')

	nok = 0
	total = 0
	wrong = dictops.dict_of_accums()
	k = None

	priorchance = prior(training, testing)
	maxcorrect = max_correct(training, testing)
	classifiers = builder(training, n_per_dim*dim,
				modelchoice=modelchoice,
				trainingset_name=trainingset_name)
	for qc in classifiers:
		failures = qc.list_wrong_classifications(testing)
		qc.add_info('correct', len(testing)-len(failures))
		qc.add_info('Ntests', len(testing))
		_addwrong(wrong, failures)
		if classout is not None:
			classout.extend( _doclass(qc, testing, verbose=verbose) )
		nok += len(failures)
		total += len(testing)
	pcorrect = float(total-nok)/float(total)

	if priorchance < maxcorrect:
		k = (pcorrect-priorchance)/(maxcorrect-priorchance)
	print '#NOK=', nok, total, float(nok)/total

	summary = {'nok': nok, 'total': total,
			'Pcorrect': float(total-nok)/float(total),
			'Chance': priorchance,
			'Pperfect': maxcorrect
			}
	if k is not None:
		summary['K'] = k

	return (summary, classifiers, wrong)



def compute_self_class(d, coverage=None, ftest=None,
		modelchoice=None, n_per_dim=None,
		builder=None, classout=None, verbose=True):

	"""modelchoice here is expected to take one argument-- the data."""

	dim = get_dim(d)

	if dim <= 0:
		die.die('zero dimensional data')
	if len(d) == 0:
		die.die('No data to classify.')
	Ntry = int(round(coverage/ftest))
	bdata = bluedata(d)
	nok = 0; total = 0;
	wrong = dictops.dict_of_accums()
	out = []; k = []; pch = []; pmx = []; pcrct = []
	for tr in range(Ntry):
		print '# Building'
		testing, training = bdata.split(ftest*len(bdata))

		# Modelchoice is a closure, and evaluation happens here:
		if modelchoice is not None:
			c_modelchoice = modelchoice(training)
		else:
			c_modelchoice = None
		tsum, classifiers, twr = compute_cross_class(training, testing, c_modelchoice,
							n_per_dim, builder, classout,
							trainingset_name = tr,
							verbose=verbose
							)

		pmx.append( tsum['Pperfect'] )
		pch.append( tsum['Chance'] )
		total += tsum['total']
		nok += tsum['nok']
		pcrct.append( 1.0-float(tsum['nok'])/tsum['total'] )
		_addwrongD(wrong, twr)
		out.extend(classifiers)
		if 'K' in tsum:
			k.append( tsum['K'] )

	summary = {'nok': nok, 'total': total}

	try:
		summary['K'], ksigma = gpkavg.avg(k, None, 0.0001)
		summary['KSigma'] = ksigma
	except ValueError, x:
		die.info('%s: K' % str(x))

	try:
		summary['Chance'], chsigma = gpkavg.avg(pch, None, 0.0001)
		summary['ChSigma'] = chsigma
	except ValueError, x:
		die.info('%s: PChance' % str(x))
	try:
		summary['Perfection'], prfsigma = gpkavg.avg(pmx, None, 0.0001)
		summary['PerfectionSigma'] = prfsigma
	except ValueError, x:
		die.info('%s: Pperfect' % str(x))

	try:
		summary['Pcorrect'], psigma = gpkavg.avg(pcrct, None, 0.0001)
		summary['PSigma'] = psigma
	except ValueError, x:
		die.info('%s: Pcorrect' % str(x))


	return (summary, out, wrong)





def default_grouper(x):
	return x.uid.split('/')[0]


def group_data(d, gr):
	o = dictops.dict_of_lists()
	for t in d:
		o.add(gr(t), t)
	return o




def compute_group_class(dg, modelchoice=None, n_per_dim=None,
		builder=None, classout=None,
		grouper=default_grouper,
		verbose=True):
	"""This function makes sure that the training set and testing set
	come from different groups.   The 'grouper' returns a group
	name, when given a datum.   Modelchoice is expected to take
	one argument, the training set.
	"""

	dim = get_dim(dg)

	if dim <= 0:
		die.die('zero dimensional data')
	if len(dg) == 0:
		die.die('No data to classify.')

	nok = 0; total = 0;
	wrong = dictops.dict_of_lists()
	out = []; k = []; pch = []; pcrct = []; pmx = []
	groups = group_data(dg, grouper)
	for (tgn, testing) in groups.items():
		training = []
		for (tmpgn, tmpg) in groups.items():
			if tmpgn != tgn:
				training.extend( tmpg )
		print '# Building'

		# Modelchoice is a closure, and evaluation happens here:
		if modelchoice is not None:
			c_modelchoice = modelchoice(training)
		else:
			c_modelchoice = None
		tsum, classifiers, twr = compute_cross_class(training, testing,
							c_modelchoice, n_per_dim,
							builder, classout,
							trainingset_name = tgn,
							verbose=verbose
							)

		pmx.append( tsum['Pperfect'] )
		pch.append( tsum['Chance'] )
		total += tsum['total']
		nok += tsum['nok']
		pcrct.append( 1.0-float(tsum['nok'])/tsum['total'] )
		_addwrongD(wrong, twr)
		out.extend(classifiers)
		if 'K' in tsum:
			k.append( tsum['K'] )

	summary = {'nok': nok, 'total': total}

	try:
		summary['K'], ksigma = gpkavg.avg(k, None, 0.0001)
		summary['KSigma'] = ksigma
	except ValueError, x:
		die.info('%s: K' % str(x))

	try:
		summary['Chance'], chsigma = gpkavg.avg(pch, None, 0.0001)
		summary['ChSigma'] = chsigma
	except ValueError, x:
		die.info('%s: PChance' % str(x))
	try:
		summary['Perfection'], prfsigma = gpkavg.avg(pmx, None, 0.0001)
		summary['PerfectionSigma'] = prfsigma
	except ValueError, x:
		die.info('%s: Pperfect' % str(x))

	try:
		summary['Pcorrect'], psigma = gpkavg.avg(pcrct, None, 0.0001)
		summary['PSigma'] = psigma
	except ValueError, x:
		die.info('%s: Pcorrect' % str(x))

	return (summary, out, wrong)





class model_template(object):
	# def __init__(self):
		# pass
	# __init__.g_implements = g_implements.varargs
	# __slots__ = []

	def logp(self, datum):
		pass


class qmodel(object):
	__slots__ = ['_mu', '_invsigma', '_bias']

	def __init__(self, mu, invsigma, offset):
			# Copy the data, so there is no risk of
			# it changing underneath you...
		self._mu = Num.array(mu, Num.Float, copy=True)
		self._invsigma = Num.array(invsigma, Num.Float, copy=True)
		assert len(self._mu.shape) == 1
		assert self._invsigma.shape == (self._mu.shape * 2), "Shapes must match: %s vs %s" % (str(self._invsigma.shape), str(self._mu.shape))
		# The above is a duplication of a tuple, not multiplication.
		self._bias = offset

	def logp(self, datum):
		delta = datum - self._mu
		parab = gpkmisc.qform(delta, self._invsigma)
		return -parab/2.0 + self._bias

	def mu(self):
		return self._mu

	def invsigma(self):
		return self._invsigma

	def bias(self):
		return self._bias

	def tochunk(self, chunkwriter):
		chunkwriter.groupstart('quadratic_class_model', b=1 )
		chunkwriter.comment('Offset:')
		chunkwriter.write_float(self.bias())
		chunkwriter.comment('Mu:')
		chunkwriter.write_NumArray(self.mu(), b=1)
		chunkwriter.comment('Inverse(covariance):')
		chunkwriter.write_NumArray(self.invsigma(), b=1)
		chunkwriter.groupend()



	def __str__(self):
		return '<model: mu=%s invsigma=%s bias=%g>' % (
				str(self.mu()), str(self.invsigma()),
				self.bias())


def _qmodel_fromchunk(chunk):
	"""This the inverse of qmodel.tochunk(),
	except the group start is already read.
	"""
	offset = chunk.read_float()
	mu = chunk.read_NumArray()
	if len(mu.shape) != 1:
		raise chunkio.BadFileFormat, 'mu must be 1-d array'
	invsigma = chunk.read_NumArray()
	if len(invsigma.shape) != 2:
		raise chunkio.BadFileFormat, 'invsigma must be 1-d array'
	if invsigma.shape != mu.shape*2:
		raise chunkio.BadFileFormat, 'sizes do not match'
	return qmodel(mu, invsigma, offset)


def _lmodel_fromchunk(chunk):
	"""This the inverse of lmodel.tochunk(),
	except the group start is already read.
	"""
	offset = chunk.read_float()
	direction = chunk.read_NumArray()
	if len(direction.shape) != 1:
		raise chunkio.BadFileFormat, 'direction must be 1-d array'
	ref = chunk.read_NumArray()
	if len(direction.shape) != 1:
		raise chunkio.BadFileFormat, 'reference_point must be 1-d array'
	if direction.shape != ref.shape:
		raise chunkio.BadFileFormat, 'sizes do not match'
	return lmodel(direction, offset, ref)


def model_fromchunk(chunk):
	"""A factory function to read in an arbitrary classifier model."""
	tmp = chunk.groupstart()
	if tmp == 'quadratic_class_model':
		rv = _qmodel_fromchunk(chunk)
	elif tmp == 'linear_class_description':
		rv = _lmodel_fromchunk(chunk)
	else:
		raise chunkio.BadFileFormat, 'Unknown type of classifier model: %s' % tmp
	chunk.groupend()
	return rv


_qzmodel_cache = {}
def qzmodel(ndim):
	global _qzmodel_cache
	if ndim in _qzmodel_cache:
		return _qzmodel_cache[ndim]
	zmu = Num.zeros((ndim,), Num.Float)
	zis = Num.zeros((ndim, ndim), Num.Float)
	zm = qmodel(zmu, zis, 0.0)
	if len(_qzmodel_cache) > 20:
		_qzmodel_cache.popitem()
	_qzmodel_cache[ndim] = zm
	return zm




class lmodel(object):
	__slots__ = ['direction', 'bias', 'reference']

	def __init__(self, direction, offset, reference_pt):
			# Copy direction, so no chance of it changing
			# due to external causes...
		self.direction = Num.array(direction, Num.Float, copy=True)
		self.bias = offset
		self.reference = reference_pt

	def logp(self, datum):
		# print "\t bias=", self.bias
		# print "\tdir=", self.direction
		# print "\tdatum=", datum
		# print "\treference=", self.reference
		# print '\to=', self.bias+Num.dot(datum-self.reference, self.direction)
		return self.bias + Num.dot(datum-self.reference, self.direction)

	def __str__(self):
		return '<model: %s ref=%s bias=%g>' % (str(self.direction), str(self.reference), self.bias)

	def tochunk(self, chunkwriter):
		chunkwriter.groupstart('linear_class_description', b=1 )
		chunkwriter.comment('Offset:')
		chunkwriter.write_float(self.bias)
		chunkwriter.comment('dir:')
		chunkwriter.write_NumArray(self.direction, b=1)
		chunkwriter.comment('reference_pt:')
		chunkwriter.write_NumArray(self.reference, b=1)
		chunkwriter.groupend()



_lzmodel_cache = {}
def lzmodel(ndim):
	if ndim in _lzmodel_cache:
		return _lzmodel_cache[ndim]
	z = Num.zeros((ndim,), Num.Float)
	zm = lmodel(z, 0.0, z)
	if len(_lzmodel_cache) > 50:
		_lzmodel_cache.popitem()
	_lzmodel_cache[ndim] = zm
	return zm




class classifier_desc(object):
	def __init__(self, list_of_classes, fvdim, evaluator=None, ftrim=None):
		assert type(fvdim) == types.IntType
		assert fvdim > 0
		self.ndim = fvdim

		# Make sure that list_of_classes is iterable
		self.c = list(list_of_classes)

		self.nc = len(self.c)
		self.evaluator = evaluator
		self.ftrim = ftrim
		self.cmap = {}
		for (i, c) in enumerate(self.c):
			self.cmap[c] = i

	def np(self):
		"""The number of parameters required to define one class."""
		raise RuntimeError, "Virtual Function."

	def npt(self):
		"""The number of parameters required to define the classifier."""
		raise RuntimeError, "Virtual Function."

	def unpack(self, prmvec, trainingset_name=None, uid=None):
		"""Produce a classifier from a parameter vector."""
		raise RuntimeError, "Virtual Function."

	def start(self, data):
		"""Starting position for Markov Chain Monte Carlo."""
		raise RuntimeError, "Virtual Function."

	def __str__(self):
		return "<classifier_desc: ndim=%d nc=%d cmap=%s>" % (self.ndim, self.nc, self.cmap)
	__repr__ = __str__


def _logp(x, (data, cdesc)):
	assert isinstance(cdesc, classifier_desc)
	cl = cdesc.unpack(x)
	assert isinstance(cl, classifier)
	o1 = evaluate_Bayes(cl, data)
	# The following little term prevents infinitely  strong
	# classifiers, in case the classes are completely
	# seperable.   Basically, it's there to prevent
	# overflows.
	o = -o1 - (TGT/o1)
	if D:
		print 'cl=', cl, 'o=', o1, o
	return o



def forest_build(data, N, modelchoice=None, trainingset_name=None):
	start, V = modelchoice.start(data)
	x = mcmc.bootstepper(_logp, start, V, c=(data, modelchoice))
	mcmch = mcmc_helper.stepper(x)
	nsteps = mcmch.run_to_bottom()
	if nsteps > 100:
		print '#NSTEPS:', nsteps
	o = []
	for i in range(N):
		mcmch.run_to_ergodic(ERGCOVER/float(N))
		tmp = modelchoice.unpack( x.current().prms(),
					trainingset_name=trainingset_name,
					uid=hash((trainingset_name,i)) )
		o.append( (tmp.evaluate(data), tmp) )
	o.sort()
	m = int(round(math.sqrt(N)))
	return [ cl for (evaluation,cl) in o[:m] ]



class classifier(object):
	def __init__(self, typename, models, info = None,
			trainingset_name=None, uid=None):
		self.typename = typename
		assert type(models) == types.DictType
		self.class_models = models
		# Info is not part of the internal operation of the classifier,
		# but it is stuff that is important to write out.
		if info is not None:
			self.info = info.copy()
		else:
			self.info = {}
		self.info['trainingset'] = trainingset_name
		self.info['Cuid'] = uid


	def add_model(self, classname, model):
		"""Add a class to an existing classifier."""
		g_implements.check(model, model_template)
		assert type(classname) == types.StringType
		self.class_models[classname] = model


	def list_classes(self):
		return self.class_models.keys()


	def add_info(self, k, v):
		self.info[k] = v

	def P(self, datum, whichclass=None):
		"""Determine the probability of being in each class.
		If whichclass=None, then it returns a list of
		tuples [(classname,P), ...] for all classes.
		Otherwise, it returns the probability of the
		desired class.
		"""
		EXPLIM = -30
		HUGE = 1e30
		lP = []
		cnames = self.class_models.keys()
		# print "Datum=", datum
		bgst = -HUGE
		for qp in self.class_models.itervalues():
			tmp = qp.logp(datum.value)
			lP.append(tmp)
			if tmp > bgst:
				bgst = tmp
		Psum = 0.0
		P = []
		for lp in lP:
			tmp = math.exp(max(lp-bgst, EXPLIM))
			P.append( tmp )
			Psum += tmp
		if whichclass is None:
			return [(cn,P/Psum) for (cn,P) in zip(cnames, P) ]
		return P[cnames.index(whichclass)]/Psum


	def bestc(self, datum):
		"""Determine the best class for a datum."""
		HUGE = 1e30
		# print "Datum=", datum
		bgst = -HUGE
		bgc = None
		for (c, qp) in self.class_models.iteritems():
			tmp = qp.logp(datum.value)
			if tmp > bgst:
				bgst = tmp
				bgc = c
		return bgc


	def __str__(self):
		o = [ '<classifier:' ]
		for (c, q) in self.class_models.items():
			o.append('%s: %s' % (c, str(q)))
		o.append( '>' )
		return '\n'.join(o)

	__repr__ = __str__


	def list_wrong_classifications(self, classdata):
		nok = []
		for datum in classdata:
			bestc = self.bestc(datum)
			if bestc != datum.classid:
				nok.append(datum.uid)
		return nok




	def writer(self, dcw):
		dcw.groupstart(self.typename)
		dcw.comment('Classifier:')
		dcw.write_dict_of(self.class_models,
				lambda dcw, ac: ac.tochunk(dcw), b=1)
		info = self.info.copy()
		dcw.write_dict(info)
		dcw.groupend()


def read_classifier(chunk):
	typename = chunk.groupstart()
	models = chunk.read_dict_of(model_fromchunk)
	info = chunk.read_dict(str)
	chunk.groupend()
	return classifier(typename, models, info=info,
			trainingset_name=info.get('trainingset', None),
			uid=info.get('Cuid', None)
			)


class forest_mixin(object):
	def __init__(self, cdesc):
		self.cdesc = cdesc

	def evaluate(self, data):
		if self.cdesc.evaluator is None:
			return evaluate_Bayes(self, data)
		return self.cdesc.evaluator(self, data)


def evaluate_match(self, data):
	"""This can be passed into a classifier descriptor as
	the evaluate argument.
	It returns the number of exact matches between the classified
	data and the input, true classification.
	"""
	# print 'classifer=', self
	o = 0
	for datum in data:
		sbc = self.bestc(datum)
		# print 'sbc(', datum, ')=', sbc, 'dc=', datum.classid, 'P=', self.P(datum)
		if sbc != datum.classid:
			o += 1
	return o



class evaluate_match_w_rare:
	__doc__ = """This is called in the same way as evaluate_match
		or evaluate_Bayes.  It pretends to be a function,
		except that you can weight the values of different
		classes when you construct the class.
		"""
	__name__ = 'evaluate_match_w_rare'

	def __init__(self):
		self.weights = None
	
	def _compute_weights(self, data):
		classcounts = dictops.dict_of_accums()
		for datum in data:
			classcounts.add(datum.classid, 1)

		ctotal = 0
		for counts in classcounts.values():
			ctotal += counts
		ctotal = float(ctotal)
		weights = {}
		for (classname, counts) in classcounts.items():
			weights[classname] = -math.log(float(counts)/ctotal)
		return weights
	

	def __call__(self, clssfr, data):
		if self.weights is None:
			self.weights = self._compute_weights(data)

		o = 0
		for datum in data:
			if clssfr.bestc(datum) != datum.classid:
				o += self.weights[datum.classid]
		return o




def evaluate_Bayes(self, data):
	"""Below, we assume that some of the
	data in each class are dubious, and should
	be ignored if they are sufficiently improbable.
	We modify the probability scores of data that is
	among the worst (self.cdesc.ftrim[0] fraction),
	and limit those scores to be no larger than
	self.cdesc.ftrim[1] larger than the best score.
	This lets you limit by score or limit by fraction
	or any mixture in between.   If self.cdesc.ftrim is None,
	then no limiting or trimming is done.
	"""
	if self.cdesc.ftrim is None:
		o = 0.0
		for datum in data:
			o -= math.log(self.P(datum, datum.classid))
		return o
	# Below, we assume a fraction trimfrac of the
	# data in each class are dubious, and should
	# be ignored if they are sufficiently improbable.
	trimfrac, trimlevel = self.cdesc.ftrim
	assert trimfrac>=0 and trimfrac<=1.0
	o = dictops.dict_of_lists()
	for datum in data:
		sb = -math.log(self.P(datum, datum.classid))
		assert isinstance(datum, datum_tr), "Use ftrim only with training set!"
		o.add(datum.classid, sb)
	oo = 0.0
	for classid, scorelist in o.items():
		scorelist.sort()
		n = int(round((1.0-trimfrac)*(len(scorelist)-1)))
		for t1 in scorelist[:n]:
			oo += t1
		tcut = scorelist[0] + trimlevel
		for t2 in scorelist[n:]:
			if D and t1 < t2:
				print 'Trimming %g %g in %s:%s' % (t1, t2, classid, scorelist)
			oo += min(tcut, t2)
	return oo


def classifier_writer(dcw, a_classifier):
	a_classifier.writer(dcw)



def default_writer(summary, out, classout, wrong):
	"""Out needs to be a list, not an iterator, because we use it twice."""

	dc = chunkio.datachunk_w( open("classes.chunk", "w") )
	dc.comment('Header:')
	header = summary.copy()
	ctype = None
	for cl in out:
		if ctype is None:
			ctype = cl.typename
		else:
			assert cl.typename == ctype, 'Cannot handle mixtures of different classifiers.'
	header['classifier_type'] = ctype
	dc.write_dict( header )
	dc.comment('classifiers:')
	dc.write_array_of(out, classifier_writer, b=1)
	out = None
	dc = None
	for (uid, nfailures) in wrong.items():
		print 'WRONG', nfailures, summary['total'], uid



def list_classes(data):
	"""List the names of the classes in a dataset,
	with the most populus classes first.
	"""
	cids = dictops.dict_of_accums()
	for datum in data:
		assert type(datum.classid) == types.StringType
		cids.add(datum.classid, 1)

	cn = [(-n, cid) for (cid, n) in cids.items() ]
	cn.sort()
	return [ cid for (n, cid) in cn ]




def read_fiat(fd):
	h, d, c = fiatio.read(fd)
	for t in d:
		if 'V' in t:
			t['V'] = chunkio.stringchunk(t['V']).read_NumArray()
		if 'P' in t:
			t['P'] = chunkio.stringchunk(t['P']).read_dict(float)
	return (h, d, c)



def name_of_evaluator(e):
	"""Used to get the name of an evaluator, to write it
	to a file header.
	"""
	if e is None:
		return evaluate_Bayes.__name__
	elif hasattr(e, '__name__'):
		return e.__name__
	else:
		return str(e)



def evaluator_from_name(nm):
	if nm is None or nm == 'Bayes' or nm == 'evaluate_Bayes':
		return evaluate_Bayes
	elif nm == 'match' or nm == 'evaluate_match':
		return evaluate_match
	elif nm == 'match_w_rare' or nm == 'evaluate_match_w_rare':
		return evaluate_match_w_rare()
	else:
		die.die('Bad name for evaluator: %s' % nm)



#!/usr/bin/env python

"""A classifier that assumes that log(P) is a quadratic form.
"""


import Num
import die
import fiatio
import q_classifier_r as Q
import g_closure
import gpkmisc


COVERAGE = 3
FTEST = 0.25
N_PER_DIM = 10


class qd_classifier_desc(Q.classifier_desc):
	def __init__(self, data=None, evaluator=None, models=None, ftrim=None):
		if data is not None:
			classes = Q.list_classes(data)
			fvdim = Q.get_dim(data)	# length of the feature vector
			# print 'LIST_CLASSES'
		elif models is not None:
			classes = models.keys()
			fvdim = models.values()[0].dir.shape[0]
			# print 'MODELS.KEYS'
		Q.classifier_desc.__init__(self, classes, fvdim, models, ftrim=ftrim)

	def np(self):
		return self.ndim + 1 + (self.ndim*(self.ndim+1))/2

	def npt(self):
		return (self.nc-1) * self.np()

	def _unpm(self, prmvec):
		"""Unpack a square, symmetric matrix."""
		nd = self.ndim
		assert len(prmvec) >= (nd*(nd+1))/2
		invsigma = Num.zeros((self.ndim, self.ndim), Num.Float)
		for k in range(nd):
			tmp = prmvec[:nd-k]
			prmvec = prmvec[nd-k:]
			invsigma[k, k:] = tmp
			invsigma[k:, k] = tmp
		return (prmvec, invsigma)


	def unpack(self, prmvec, trainingset_name=None, uid=None):
		"""Produce a classifier from a parameter vector."""
		m = self.np()
		nd = self.ndim
		assert len(prmvec) == self.npt()

		# print 'UNPACK'
		o = qd_classifier(self, trainingset_name=trainingset_name, uid=uid)
		for (i, c) in enumerate(self.c):
			# print "im", i, m
			if i >= self.nc-1:
				break	# Last one is the zero model.
			pc = prmvec[i*m: (i+1)*m]
			offset = pc[0]
			pc = pc[1:]
			mu = pc[:nd]
			npc, invsigma = self._unpm(pc[nd:])
			assert len(npc) == 0
			o.add_model( c, Q.qmodel( mu, invsigma, offset))
		o.add_model( c, Q.qzmodel(self.ndim))
		# print '#UP:', o
		return o


	def start(self, data):
		"""Starting position for Markov Chain Monte Carlo."""
		m = self.np()
		data_in_class = {}
		all_data_vecs = []
		for c in self.c:
			data_in_class[c] = []
		for datum in data:
			data_in_class[datum.classid].append( datum.value )
			all_data_vecs.append( datum.value )
		amu = gpkmisc.median_across( all_data_vecs )
		mu = {}
		for c in self.c:
			assert len(data_in_class[c]) > 0
			mu[c] = gpkmisc.median_across(data_in_class[c])
			# print '#Mu[', c, ']=', mu[c]

		if len(data) < 2:
			cov = Num.identity(amu.shape[0], Num.Float)
		else:
			cov = Num.zeros((amu.shape[0], amu.shape[0]), Num.Float)
			for datum in data:
				# print '#dval', datum.value
				tmp = datum.value - amu
				# print '#tmp', c, tmp
				cov += Num.outerproduct(tmp, tmp)
			cov /= len(data)
		# print '#cov=', cov
		tr = Num.sum(Num.diagonal(cov))
		trf = 0.01*tr/cov.shape[0]
		# print '# tr=', tr, trf
		for i in range(cov.shape[0]):
			cov[i,i] += trf

		invsigma = Num.LA.inv(cov)
		# Use Mu!
		# print '#cov+ =', cov
		npt = self.npt()
		prmvec = Num.zeros((npt,), Num.Float)
		V = Num.zeros((npt, npt), Num.Float)
		j = 0
		assert len(self.c) == self.nc
		for (i, c) in enumerate(self.c):
			if i >= self.nc-1:
				break	# Last one is the zero classifier
			pc = prmvec[i*m: (i+1)*m]
			Vc = V[i*m:(i+1)*m, i*m:(i+1)*m]
			pc[0] = 0.0	# offset
			Vc[0,0] = 1.0
			j = self.ndim + 1
			pc[1:j] = mu[c]
			Vc[1:j, 1:j] = cov
			for k in range(self.ndim):
				# print "pc:", pc
				# print "invsigma:", invsigma
				w = self.ndim-k
				# print "j", j, w, j+w, c, k
				pc[j:j+w] = invsigma[k, k:k+w]
				# print "pc2:", pc
				j += w
			for jj in range(self.ndim+1, j):
				Vc[jj, jj] = pc[jj]**2
			# print "pcf:", pc
			assert j == self.np()

		check = self.unpack(prmvec)
		# print '#Check=', check	# Looking at the last term in self.c:
		for (i, c) in enumerate(self.c):
			if i < self.nc-1:
				# print '# CC=', c
				# print '# \t check.cm[c]=', check.class_models[c]
				assert check.class_models[c].bias() == 0.0
				assert Num.sum(Num.absolute(check.class_models[c].mu()-mu[c])) < 0.001
				assert Num.sum(Num.ravel(Num.absolute(check.class_models[c].invsigma()-invsigma))) < 0.001
			else:
				assert check.class_models[c].bias() == 0.0
				assert Num.sum(Num.absolute(check.class_models[c].mu())) < 0.001
				assert Num.sum(Num.ravel(Num.absolute(check.class_models[c].invsigma()))) < 0.001
		# print '#prmvec=', prmvec
		# print '#V=', V
		return (prmvec, V)

	def typename(self):
		return 'quadratic_discriminant_classifier'






class qd_classifier(Q.classifier, Q.forest_mixin):
	def __init__(self, cdesc=None, models=None, trainingset_name=None, uid=None):
		if cdesc is None:
			cdesc = qd_classifier_desc(models=models)
			
		if models is None:
			models = {}
		Q.classifier.__init__(self, cdesc.typename(),
					models, trainingset_name=trainingset_name,
					uid=uid)
		Q.forest_mixin.__init__(self, cdesc)






def go_auto(fd, coverage=COVERAGE, ftest=FTEST, n_per_dim = N_PER_DIM,
			ftrim=None, evnm=None):
	d = Q.read_data(fd)
	modelchoice = g_closure.Closure(qd_classifier_desc, g_closure.NotYet,
					evaluator=Q.evaluator_from_name(evnm),
					ftrim=ftrim)
	classout = fiatio.writer(open('classified.fiat', 'w'))
	classout.header('evaluator', evnm)
	classout.header('ftrim', ftrim)
	summary, out, wrong = Q.compute_self_class(d, coverage=coverage, ftest=ftest,
						n_per_dim = n_per_dim,
						modelchoice = modelchoice,
						builder=Q.forest_build,
						classout=classout)
	Q.default_writer(summary, out, classout, wrong)






def test_build3(classifier_choice):
	print 'TEST_BUILD3:'
	data = [
		Q.datum_tr([1.0], 'A'), Q.datum_tr([-0.01], 'B'), Q.datum_tr([2.01], 'C'),
		Q.datum_tr([1.02], 'A'), Q.datum_tr([-0.02], 'B'), Q.datum_tr([2.02], 'C'),
		Q.datum_tr([1.01], 'A'), Q.datum_tr([0.0], 'B'), Q.datum_tr([2.0], 'C'),
		Q.datum_tr([0.98], 'A'), Q.datum_tr([0.02], 'B'), Q.datum_tr([1.99], 'C'),
		Q.datum_tr([0.99], 'A'), Q.datum_tr([0.01], 'B'), Q.datum_tr([1.98], 'C'),
		]
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=None)
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = 0
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		if eval < 1:
			ok = 1
			break
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")



def test_build1q(classifier_choice):
	import random
	print 'TEST_BUILD1q:'
	data = []
	for i in range(9):
		data.append( Q.datum_tr([random.random()], 'A'))
		data.append( Q.datum_tr([random.random()+2.0], 'B'))
		data.append( Q.datum_tr([random.random()+4.0], 'A'))
	for datum in data:
		print 'd=', datum
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=None, ftrim=None)
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = 0
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		if eval < 1:
			ok = 1
			break
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")



def test_build1tq(classifier_choice):
	import random
	print 'TEST_BUILD1tq:'
	data = [ Q.datum_tr([-10.0], 'B') ]	# One spurious point to be trimmed off.
	for i in range(30):
		data.append( Q.datum_tr([random.random()], 'A'))
		data.append( Q.datum_tr([random.random()+2.0], 'B'))
		data.append( Q.datum_tr([random.random()+4.0], 'A'))
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=None, ftrim=(0.1, 6))
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = 0
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		if eval < 2:
			ok = 1
			break
		else:
			print 'eval=', eval
			print 'cdef=', cdef
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")




def test():
	from l_classifier import test_build1, test_build32, test_build2
	from l_classifier import test_2_bias, test_4_2, test_build1t
	from l_classifier import test_2_scale
	cc = qd_classifier_desc
	test_build1(cc)
	test_build1q(cc)
	# test_build1t(cc)
	test_build1tq(cc)
	test_build32(cc)
	test_build2(cc)
	test_2_scale(cc)
	test_build3(cc)
	# test_var()
	test_2_bias(cc)
	test_4_2(cc)


PSYCO = 1

if __name__ == '__main__':
	import sys

	if PSYCO:
		try:
			import psyco
			psyco.full(memory=10000)
		except ImportError:
			pass

	ftrim = None
	arglist = sys.argv[1:]
	evnm = None
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-coverage':
			COVERAGE = float(arglist.pop(0))
		elif arg == '-nperdim':
			N_PER_DIM = int(arglist.pop(0))
		elif arg == '-evaluator':
			evnm = arglist.pop(0)
		elif arg == '-ftest':
			FTEST = float(arglist.pop(0))
			assert 0.0 < FTEST < 1.0
		elif arg == '-c':
			ftrim = float(arglist.pop(0))
			assert 0.0 <= ftrim < 0.5
		elif arg == '-D':
			Q.D = 1
		elif arg == '-test':
			test()
			break
		else:
			die.die('Unrecognized argument: %s' % arg)
	go_auto(sys.stdin, coverage=COVERAGE, n_per_dim = N_PER_DIM,
		ftrim=ftrim, evnm=evnm)



#!/usr/bin/env python

import l_classifier as L
import die
import q_classifier_r as Q
import fiatio


N_PER_DIM = 6



def go_cross(trfd, tsfd, n_per_dim = N_PER_DIM, ftrim=0.0):
	trd = Q.read_data(trfd)
	tsd = Q.read_data(tsfd)
	modelchoice = L.l_classifier_desc(trd+tsd, userank=1, ftrim=ftrim)
	classout = fiatio.writer(open('classified.fiat', 'w'))
	classout.header('userank', modelchoice.userank)
	classout.header('ftrim', modelchoice.ftrim)
	for (i, c) in enumerate(modelchoice.c):
		classout.header('Class%0d' % i, c)

	summary, out, wrong = Q.compute_cross_class(trd, tsd,
						modelchoice = modelchoice,
						n_per_dim = n_per_dim,
						builder=Q.forest_build,
						classout=classout)
	L.default_writer(summary, out, classout, wrong, modelchoice)



if __name__ == '__main__':
	import sys

	try:
		import psyco
		psyco.full(memory=10000)
	except ImportError:
		pass

	arglist = sys.argv[1:]
	ftrim = None
	while len(arglist)>0 and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-c':
			ftrim = float(arglist.pop(0))
		elif arg == '-nperdim':
			N_PER_DIM = int(arglist.pop(0))
		else:
			die.die('Unrecognized argument: %s' % arg)
	assert len(arglist) == 2
	trfd = open(arglist.pop(0), 'r')
	tsfd = open(arglist.pop(0), 'r')

	L.go_cross(trfd, tsfd, n_per_dim = N_PER_DIM, ftrim=ftrim)



#!/usr/bin/env python

"""A quadratic classifier, but something about groups."""

import q_classifier_r as Q
import fiatio
import die
import g_closure
import qd_classifier as QD


N_PER_DIM = 6
COVERAGE = 3


def go_group_q(fd, n_per_dim = N_PER_DIM, ftrim=None,
		grouper=Q.default_grouper,
		evnm=None):
	d = Q.read_data(fd)

	classout = fiatio.writer(open('classified.fiat', 'w'))

	modelchoice = g_closure.Closure(QD.qd_classifier_desc, g_closure.NotYet,
					evaluator=Q.evaluator_from_name(evnm),
					ftrim=ftrim)
	classout.header('evaluator', evnm)
	classout.header('ftrim', ftrim)

	summary, out, wrong = Q.compute_group_class(
						d,
						modelchoice = modelchoice,
						n_per_dim = n_per_dim,
						builder=Q.forest_build,
						classout=classout,
						grouper=grouper)

	Q.default_writer(summary, out, classout, wrong)


def go_group_l(fd, n_per_dim = N_PER_DIM, ftrim=None,
		grouper=Q.default_grouper,
		evnm=None):
	import l_classifier
	d = Q.read_data(fd)

	classout = fiatio.writer(open('classified.fiat', 'w'))

	modelchoice = g_closure.Closure(l_classifier.l_classifier_desc, g_closure.NotYet,
					evaluator=Q.evaluator_from_name(evnm),
					ftrim=ftrim)
	classout.header('evaluator', evnm)
	classout.header('ftrim', ftrim)

	summary, out, wrong = Q.compute_group_class(
						d,
						modelchoice = modelchoice,
						n_per_dim = n_per_dim,
						builder=Q.forest_build,
						classout=classout,
						grouper=grouper)

	Q.default_writer(summary, out, classout, wrong)




if __name__ == '__main__':
	import sys

	try:
		import psyco
		psyco.full(memory=10000)
	except ImportError:
		pass

	arglist = sys.argv[1:]
	ftrim = None
	ct = 'Q'
	evnm = None
	while len(arglist)>0 and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-nperdim':
			N_PER_DIM = int(arglist.pop(0))
		elif arg == '-Q':
			ct = 'Q'
		elif arg == '-L':
			ct = 'L'
		elif arg == '-evaluator':
			evnm = arglist.pop(0)
		elif arg == '-c':
			trimf = float(arglist.pop(0))
			triml = float(arglist.pop(0))
			ftrim = (trimf, triml)
		elif arg == '-D':
			Q.D = 1
		else:
			die.die('Unrecognized argument: %s' % arg)
	assert len(arglist) == 0

	if ct == 'Q':
		go_group_q(sys.stdin, n_per_dim = N_PER_DIM, ftrim=ftrim, evnm=evnm)
	else:
		go_group_l(sys.stdin, n_per_dim = N_PER_DIM, ftrim=ftrim, evnm=evnm)



#!/usr/bin/env python

"""A classifier that assumes that P is linear in position.
This is known as a (linear) logistic discriminant analysis::

@inbook{webb:spr:logistic,
   author    = {Andrew Webb},
   title     = {Statistical Pattern Recognition},
   pages     = {124--132},
   year      = {1999},
   publisher = {Arnold},
   address   = {London, New York},
   note = {ISBN 0 340 74164  3}
}
"""


import Num
import chunkio
import die
import fiatio
import q_classifier_r as Q
import g_closure
import gpkmisc
import Numeric_gpk
import math
import random
import sys

COVERAGE = 3
FTEST = 0.25
N_PER_DIM = 10


def ref_from_data(data):
	tmp = [ datum.value for datum in data ]
	return gpkmisc.median_across(tmp)


def ref_from_models(models):
	tmp = [m.reference for m in models.values()]
	return gpkmisc.median_across(tmp)


class l_classifier_desc(Q.classifier_desc):
	def __init__(self, data=None, evaluator=None, models=None, ftrim=None):
		if data is not None:
			classes = Q.list_classes(data)
			fvdim = Q.get_dim(data)	# length of the feature vector
			reference_pt = ref_from_data(data)
			# print 'LIST_CLASSES'
		elif models is not None:
			classes = models.keys()
			fvdim = models.values()[0].direction.shape[0]
			reference_pt = ref_from_models(models)
			# print 'MODELS.KEYS'
		Q.classifier_desc.__init__(self, classes, fvdim, models, ftrim=ftrim)
		self.reference = reference_pt
		self.evaluator = evaluator


	def np(self):
		"""The number of parameters required to define each class."""
		assert self.ndim > 0
		return self.ndim + 1

	def npt(self):
		"""The total number of parameters required to define all classes."""
		assert self.nc > 1
		assert self.np() > 0
		return (self.nc-1) * self.np()

	def unpack(self, prmvec, trainingset_name=None, uid=None):
		"""Produce a classifier from a parameter vector."""
		m = self.np()
		assert len(prmvec) == self.npt()
		assert self.c	# Can't do much with zero classes.

		o = l_classifier(self, trainingset_name=trainingset_name, uid=uid)
		for (i, c) in enumerate(self.c):
			if i >= self.nc-1:
				break	# Last one is the zero model.
			# print 'ic', i, c, i*m, (i+1)*m, prmvec.shape
			pc = prmvec[i*m : (i+1)*m]
			bias = pc[0]
			prms = pc[1:]
			o.add_model( c, Q.lmodel( prms, bias, self.reference ) )
		o.add_model(c, Q.lzmodel( self.ndim ) )
		return o


	def start(self, data):
		"""Starting position for Markov Chain Monte Carlo."""
		assert self.npt() > 0
		p = Num.zeros((self.npt(),), Num.Float)
		v = Num.identity(self.npt(), Num.Float)
		var = Numeric_gpk.vec_variance( [datum.value for datum in data] )
		print 'data[0].value=', data[0].value
		print 'START: var=', var
		assert var.shape == (self.ndim,)
		assert self.npt() % self.np() == 0
		for i in range(self.npt()):
			snp = self.np()
			isnp = i % snp
			if isnp == 0:
				sig = 1.0
			else:
				sig = math.sqrt(var[isnp-1])
			p[i] = random.normalvariate(0.0, 1.0/sig)
			v[i,i] = sig**-2
		print 'p=', p
		print 'v=', v
		return (p, v)

	def typename(self):
		return 'linear_discriminant_classifier'






class l_classifier(Q.classifier, Q.forest_mixin):
	def __init__(self, cdesc=None, models=None, trainingset_name=None, uid=None):
		if cdesc is None:
			cdesc = l_classifier_desc(models=models)
		if models is None:
			models = {}
		Q.classifier.__init__(self, cdesc.typename(), models,
					trainingset_name=trainingset_name, uid=uid)
		Q.forest_mixin.__init__(self, cdesc)
	





def go_auto(fd, coverage=COVERAGE, ftest=FTEST, n_per_dim = N_PER_DIM,
			ftrim=None, evnm=None, verbose=True):
	data = Q.read_data(fd)
	modelchoice = g_closure.Closure(l_classifier_desc, g_closure.NotYet,
					evaluator=Q.evaluator_from_name(evnm),
					ftrim=ftrim)
	classout = fiatio.writer(open('classified.fiat', 'w'))
	classout.header('evaluator', evnm)
	classout.header('ftrim', ftrim)
	summary, out, wrong = Q.compute_self_class(data, coverage=coverage, ftest=ftest,
						n_per_dim = n_per_dim,
						modelchoice = modelchoice,
						builder=Q.forest_build,
						classout=classout,
						verbose=verbose)
	Q.default_writer(summary, out, classout, wrong)




def classifier_reader(dc):
	"""Read the classifier definitions produced by classifier_writer()."""
	def model_reader(dc):
		if dc.groupstart() != 'linear_class_description':
			raise chunkio.BadFileFormat, "Wrong type of class description"
		bias = dc.read_float()
		direction = dc.read_NumArray()
		if len(direction.shape) != 1:
			raise chunkio.BadFileFormat, "Bad shape direction"
		reference_pt = dc.read_NumArray()
		dc.groupend()
		q = Q.model(direction, bias, reference_pt)
		return q

	if dc.groupstart() != 'linear_discriminant_classifier':
		raise chunkio.BadFileFormat, "Missing outer group"

	models = dc.read_dict_of(model_reader)
	info = dc.read_dict(lambda x:x)
	dc.groupend()
	return l_classifier( cdesc=None, models=models, trainingset_name=info['trainingset_name'], uid=info['uid'])


def read_classes(fd):
	dc = chunkio.datachunk(fd)
	hdr = dc.read_dict(float)
	classes = dc.read_array_of(classifier_reader)
	return (hdr, classes)




def test_build1(classifier_choice):
	print 'TEST_BUILD1:'
	data = [ Q.datum_tr([1.0], 'A'), Q.datum_tr([0.0], 'B') ]
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=Q.evaluate_match)
	assert isinstance(modelchoice, Q.classifier_desc)
	print 'modelchoice=', modelchoice
	print 'data=', data
	ok = 0
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		print 'cdef=', cdef
		eigenvalue = Q.evaluate_match(cdef, data)
		print 'eval=', eigenvalue
		if eigenvalue < 1:
			ok = 1
			break
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")



def test_build1t(classifier_choice):
	print 'TEST_BUILD1t:'
	# Two bad points to be trimmed off:
	data = [ Q.datum_tr([100.0], 'A'), Q.datum_tr([-101.0], 'B') ]
	# Lots of good points:
	for i in range(30):
		data.append( Q.datum_tr([random.random()], 'A'))
		data.append( Q.datum_tr([random.random()+2.0], 'B'))
	print 'data=', data
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=Q.evaluate_match, ftrim=(0.1,6))
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = 0
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		if eval < 3:
			ok = 1
			break
		else:
			print 'eval=', eval
			print 'cdef=', cdef
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")



def test_build32(classifier_choice):
	print 'TEST_BUILD32:'
	data = [
		Q.datum_tr([1.0, 0.0], 'A'),
		Q.datum_tr([-0.01, 0.0], 'B'),
		Q.datum_tr([-.01, 1.0], 'C'),
		Q.datum_tr([1.02, 0.01], 'A'),
		Q.datum_tr([-0.02, 0.01], 'B'),
		Q.datum_tr([-.02, 1.01], 'C'),
		Q.datum_tr([1.01, -0.01], 'A'),
		Q.datum_tr([0.0, -0.01], 'B'),
		Q.datum_tr([0, 0.99], 'C'),
		Q.datum_tr([0.98, 0.02], 'A'),
		Q.datum_tr([0.02, 0.0], 'B'),
		Q.datum_tr([-0.01, 0.98], 'C'),
		Q.datum_tr([0.99, 0.01], 'A'),
		Q.datum_tr([0.01, 0.03], 'B'),
		Q.datum_tr([-.02, 0.97], 'C'),
		]
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=Q.evaluate_match)
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = False
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		print cdef
		if eval < 1:
			ok = True
			break
	if not ok:
		die.die( "can't find a perfect classifier in this trivial situation!")


def test_build2(classifier_choice):
	print 'TEST_BUILD2:'
	data = [
		Q.datum_tr([1.0, 2.0, 0.0], 'A'),
		Q.datum_tr([0.0, 1.0, 0.0], 'A'),
		Q.datum_tr([1.0, 1.0, 1.0], 'A'),
		Q.datum_tr([2.0, 0.0, 0.0], 'A'),
		Q.datum_tr([2.0, 2.0, 2.0], 'A'),
		Q.datum_tr([1.0, 0.0, 1.0], 'A'),
		Q.datum_tr([0.0, -1.0, -2.0], 'B'),
		Q.datum_tr([-1.0, -1.0, -2.0], 'B'),
		Q.datum_tr([0.0, -1.0, -1.0], 'B'),
		Q.datum_tr([-2.0, -1.0, -2.0], 'B'),
		Q.datum_tr([0.0, -1.0, 0.0], 'B'),
		Q.datum_tr([0.0, 0.0, -2.0], 'B'),
		Q.datum_tr([-2.0, -2.0, -2.0], 'B'),
		Q.datum_tr([-1.0, -1.0, 0.0], 'B')
		]
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = classifier_choice(data, evaluator=Q.evaluate_match)
	assert isinstance(modelchoice, Q.classifier_desc)
	ok = False
	for cdef in Q.forest_build(data, 40, modelchoice=modelchoice):
		eval = Q.evaluate_match(cdef, data)
		if eval < 1:
			ok = True
			break
	if not ok:
		die.die("can't find a perfect classifier in this trivial situation!")



def test_4_2(classifier_choice):
	print 'TEST_4_2:'
	data = []
	for i in range(400):
		data.append( Q.datum_tr( [ random.random() ], 'a') )
		data.append( Q.datum_tr( [ random.random() ], 'b') )
		data.append( Q.datum_tr( [ 5+2*random.random() ], 'c') )
		data.append( Q.datum_tr( [ 5+2*random.random() ], 'd') )
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = g_closure.Closure(classifier_choice, g_closure.NotYet,
					evaluator=Q.evaluate_match)
	summary, out, wrong = Q.compute_self_class(data, ftest=FTEST, coverage=COVERAGE,
						modelchoice = modelchoice,
						n_per_dim = N_PER_DIM,
						builder=Q.forest_build)
	assert abs(summary['Chance'] - 0.25) < 0.1
	assert abs(summary['Pcorrect'] - 0.5) < 0.1
	assert summary['Pcorrect'] > summary['Chance']
	print 'summary:', summary



def test_2_bias(classifier_choice):
	print 'TEST_2_bias:'
	data = []
	for i in range(400):
		data.append( Q.datum_tr( [ random.random() ], 'a') )
	for i in range(40):
		data.append( Q.datum_tr( [ random.random() ], 'b') )
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = g_closure.Closure(classifier_choice, g_closure.NotYet,
					evaluator=Q.evaluate_match)
	classout = []
	summary, out, wrong = Q.compute_self_class(data, ftest=FTEST, coverage=COVERAGE,
						modelchoice = modelchoice,
						n_per_dim = N_PER_DIM,
						builder=Q.forest_build,
						classout=classout)
	print 'summary:', summary
	assert abs(summary['Chance'] - 0.9) < 0.05
	assert abs(summary['Pcorrect'] - 0.9) < 0.05
	nbc = 0
	nbt = 0
	ntot = 0
	for tmp in classout:
		if tmp['trueclass']=='b':
			nbt += 1
		if tmp['compclass']=='b':
			nbc += 1
		ntot += 1
	print 'true b=', nbt/float(ntot), 'computed b=', nbc/float(ntot)
	assert abs(nbt/float(ntot) - 0.1) < 0.03
	assert nbc < 1.3*nbt


def _t2sguts(off, f):
	o = []
	for i in range(1,8):
		o.append( random.gauss(off*i, f*i*i) )
	return o

def test_2_scale(classifier_choice):
	print 'TEST_2_scale:'
	data = []
	for i in range(300):
		data.append( Q.datum_tr( _t2sguts(0.0, 0.001), 'a') )
	for i in range(100):
		data.append( Q.datum_tr( _t2sguts(0.0, 0.001), 'b') )
	assert issubclass(classifier_choice, Q.classifier_desc)
	modelchoice = g_closure.Closure(classifier_choice, g_closure.NotYet,
					evaluator=Q.evaluate_match)
	classout = []
	summary, out, wrong = Q.compute_self_class(data, ftest=FTEST, coverage=COVERAGE,
						modelchoice = modelchoice,
						n_per_dim = N_PER_DIM,
						builder=Q.forest_build,
						classout=classout)
	print 'summary:', summary
	assert abs(summary['Chance'] - 0.75) < 0.05
	assert abs(summary['Pcorrect'] - 0.75) < 0.10
	nbc = 0
	nbt = 0
	ntot = 0
	for tmp in classout:
		if tmp['trueclass']=='b':
			nbt += 1
		if tmp['compclass']=='b':
			nbc += 1
		ntot += 1
	print 'true b=', nbt/float(ntot), 'computed b=', nbc/float(ntot)
	assert abs(nbt/float(ntot) - 0.25) < 0.10
	assert nbc < 1.3*nbt



def test_var():
	print 'TEST_var:'
	data = []
	for i in range(80):
		data.append( Q.datum_tr( [ random.gauss(0.0, 1.0), random.gauss(0.0, 1.0) ], 'a') )
	for i in range(160):
		data.append( Q.datum_tr( [ random.gauss(0.0, 1.0), random.gauss(0.0, 1.0)+0.01 ], 'b') )
	modelchoice = g_closure.Closure(l_classifier_desc, data,
					evaluator=Q.evaluate_match)
	summary, out, wrong = Q.compute_self_class(data,
					modelchoice = modelchoice,
					builder=Q.forest_build)
	print 'summary:', summary
	# assert abs(summary['Chance'] - 0.5) < 0.1
	# assert abs(summary['Pcorrect'] - 0.75) < 0.1


def calibrate_var():
	for i in range(15):
		test_var()
		sys.stdout.flush()

def test():
	cc = l_classifier_desc
	test_build1(cc)
	test_build1t(cc)
	test_build32(cc)
	test_build2(cc)
	# test_var()
	test_2_scale(cc)
	test_2_bias(cc)
	test_4_2(cc)


PSYCO = 1

if __name__ == '__main__':

	if PSYCO:
		try:
			import psyco
			psyco.full(memory=10000)
		except ImportError:
			pass

	evnm = None
	ftrim = None
	Verbose = True
	arglist = sys.argv[1:]
	while len(arglist)>0 and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-coverage':
			COVERAGE = float(arglist.pop(0))
		elif arg == '-nperdim':
			N_PER_DIM = int(arglist.pop(0))
		elif arg == '-evaluator':
			evnm = arglist.pop(0)
		elif arg == '-ftest':
			FTEST = float(arglist.pop(0))
		elif arg == '-c':
			ftrim = float(arglist.pop(0))
		elif arg == '-taciturn':
			verbose = False
		elif arg == '-D':
			Q.D = 1
		elif arg == '-test':
			test()
			sys.exit(0)
		else:
			die.die('Unrecognized argument: %s' % arg)
	go_auto(sys.stdin, coverage=COVERAGE, n_per_dim = N_PER_DIM,
		ftrim=ftrim, evnm=evnm, ftest=FTEST, verbose=Verbose)



"""Implements the creation of quadratic and diagonal quadratic
classifiers for q_classifier.py.
"""



import multivariance
import math
import q_classifier as QC
import random

HUGE=1e30

def build_classifiers(data, N, modelchoice=multivariance.quadratic):
	"""Builds a set of quadratic classifiers.
	Data is a list of datum_tr .
	It returns a list:
		[ (evaluation_result, classifier_description), ... ] .
	"""
	assert modelchoice==multivariance.quadratic	\
		or modelchoice==multivariance.diag_quadratic
	dcl = {}
	for datum in data:
		try:
			dcl[datum.classid].append(datum.value)
		except KeyError:
			dcl[datum.classid] = [ datum.value ]
	ncl = len(dcl)
	# print '# ncl=', ncl
	# nvar = 1 + int(math.ceil(math.sqrt(N)))
	nvar = 1 + int(math.ceil( min(float(3*N)**(1.0/len(dcl)), math.sqrt(N)) ))
	cprms = {}
	for (c, cd) in dcl.items():
		print '# Multivar on %s' % c
		cprmcl = multivariance.meanvar(cd, nvar, modelchoice)

		# This is effectively the Bayesean prior: classes where we have
		# seen lots of examples are more probable.
		for cprmc in cprmcl:
			cprmc.bias = math.log(len(cd))
		cprms[c] = cprmcl
		# print '\t cprmcl:', cprmcl

		assert len(cprms[c]) > 0

	classifiers = []
	# Now, we build a bunch of classifiers
	for i in range(N):
		qc = QC.classifier()
		assert len(qc.class_models) == 0
		# For each class, choose one plausible description:
		for (c, cpa) in cprms.items():
			# print 'C=', c, 'cpa=', cpa
			assert len(cpa)>1
			qc.addQuad(c, random.choice(cpa))
			# print 'qc=', qc
		classifiers.append(qc)
	# print 'classifiers=', classifiers
	return classifiers



#!/usr/bin/env python

import Num
import blue_data_selector
import die



def _mswap(sw1, sw2, m):
	"""Swap rows and columns in matrix m, according to mapping swv.
	Produces a copy of the matrix."""
	assert sw1.shape == (m.shape[0],)
	assert sw2.shape == (m.shape[1],)
	m1 = Num.take(m, sw1, axis=0)
	return Num.take(m1, sw2, axis=1)


def _lswap(swv, m):
	"""Swap labels in list m, according to mapping swv.
		Produces a new list.
		"""
	assert swv.shape == (len(m),)
	return [ m[swv[i]] for i in range(len(m)) ]



class diagfom(object):
	def __init__(self, n):
		ar = Num.arrayrange(n)
		self.weight = Num.absolute(ar[:,Num.NewAxis] - ar[Num.NewAxis,:])

	def eval(self, swv, m):
		m2 = _mswap(swv, swv, m)
		# return Num.sum(Num.ravel(m2*self.weight))
		return Num.sum(m2*self.weight)


class diagfom2(object):
	def __init__(self, n1, n2):
		a1 = Num.arrayrange(n1)
		a2 = Num.arrayrange(n2)
		self.weight = Num.absolute(a1[:,Num.NewAxis]/float(n1-1) - a2[Num.NewAxis,:]/float(n2-1))

	def eval(self, sw1, sw2, m):
		m2 = _mswap(sw1, sw2, m)
		# return Num.sum(Num.ravel(m2*self.weight))
		return Num.sum(m2*self.weight)


class blockfom(object):
	def __init__(self, n1, n2):
		pass

	def eval(self, sw1, sw2, m):
		m2 = _mswap(sw1, sw2, m)
		o = 0.0
		# o += Num.sum(Num.ravel(Num.absolute(m2[1:,:] - m2[:-1,:])))
		# o += Num.sum(Num.ravel(Num.absolute(m2[:,1:] - m2[:,:-1])))
		o += Num.sum(Num.absolute(m2[1:,:] - m2[:-1,:]))
		o += Num.sum(Num.absolute(m2[:,1:] - m2[:,:-1]))
		return o


def diag_swap_toward_minimal_fom(m, lbls, fom, maxtries):
	n = len(lbls)
	assert m.shape == (n, n)
	swv = Num.arrayrange(n)
	of = fom.eval(swv, m)
	sincelast = 0
	bds = blue_data_selector.bluedata( [ (i,j) for i in range(n) for j in range(n)] )
	if maxtries is None:
		maxtries = 2*n*n
	for (i,j) in bds:
		oswv = Num.array(swv, Num.Int, copy=True)
		swv[i], swv[j] = (swv[j], swv[i])
		nf = fom.eval(swv, m)
		if nf < of:
			of = nf
			sincelast = 0
		else:
			swv = oswv
			sincelast += 1
		if sincelast > maxtries:
			break
	return ( _mswap(swv, swv, m), _lswap(swv, swv, lbls) )


def swap_toward_minimal_fom(m, lbl1, lbl2, fom, maxtries):
	n1 = len(lbl1)
	n2 = len(lbl2)
	assert m.shape == (n1, n2)
	sw1 = Num.arrayrange(n1)
	sw2 = Num.arrayrange(n2)
	of = fom.eval(sw1, sw2, m)
	sincelast = 0
	bd1 = blue_data_selector.bluedata( [ (i,j) for i in range(n1) for j in range(n1)] )
	bd2 = blue_data_selector.bluedata( [ (i,j) for i in range(n2) for j in range(n2)] )
	if maxtries is None:
		maxtries = 2*(n1*n1 + n2*n2)
	else:
		maxtries = 2*maxtries
	while True:
		i, j = bd1.pick(1)[0]
		osw1 = Num.array(sw1, Num.Int, copy=True)
		sw1[i], sw1[j] = (sw1[j], sw1[i])
		nf = fom.eval(sw1, sw2, m)
		if nf < of:
			die.info('fom=%g->%g; sincelast=%d' % (of, nf, sincelast))
			of = nf
			sincelast = 0
		else:
			sw1 = osw1
			sincelast += 1

		i, j = bd2.pick(1)[0]
		osw2 = Num.array(sw2, Num.Int, copy=True)
		sw2[i], sw2[j] = (sw2[j], sw2[i])
		nf = fom.eval(sw1, sw2, m)
		if nf < of:
			of = nf
			die.info('fom=%g->%g; sincelast=%d' % (of, nf, sincelast))
			sincelast = 0
		else:
			sw2 = osw2
			sincelast += 1
		if sincelast > maxtries:
			break
	return ( _mswap(sw1, sw2, m), _lswap(sw1, lbl1), _lswap(sw2, lbl2) )



def swap_toward_diag(m, lbls, maxtries=None):
	fom = diagfom(len(lbls))
	return diag_swap_toward_minimal_fom(m, lbls, fom, maxtries)


def swap_toward_diag2(m, lbl1, lbl2, maxtries=None):
	fom = diagfom2(len(lbl1), len(lbl2))
	return swap_toward_minimal_fom(m, lbl1, lbl2, fom, maxtries)


def swap_toward_blocks(m, lbl1, lbl2, maxtries=None):
	fom = blockfom(len(lbl1), len(lbl2))
	return swap_toward_minimal_fom(m, lbl1, lbl2, fom, maxtries)



import Num


def wp(data, wt, percentiles):
	assert Num.alltrue(Num.greater_equal(percentiles, 0.0)), "Percentiles less than zero"
	assert Num.alltrue(Num.less_equal(percentiles, 1.0)), "Percentiles greater than one"
	data = Num.asarray(data)
	if wt is None:
		wt = Num.ones(data.shape, Num.Float)
	else:
		wt = Num.asarray(wt, Num.Float)
		assert wt.shape == data.shape
		assert Num.alltrue(Num.greater_equal(wt, 0.0)), "Not all weights are non-negative."
	assert len(wt.shape) == 1
	assert len(data.shape) == 1
	n = data.shape[0]
	i = Num.argsort(data)
	sd = Num.take(data, i, axis=0)
	sw = Num.take(wt, i, axis=0)
	aw = Num.add.accumulate(sw)
	if not aw[-1] > 0:
		raise ValueError, "Nonpositive weight sum"
	w = (aw-0.5*sw)/aw[-1]
	spots = Num.searchsorted(w, percentiles)
	o = []
	for (s, p) in zip(spots, percentiles):
		if s == 0:
			o.append(sd[0])
		elif s == n:
			o.append(sd[n-1])
		else:
			f1 = (w[s] - p)/(w[s] - w[s-1])
			f2 = (p - w[s-1])/(w[s] - w[s-1])
			assert f1>=0 and f2>=0 and f1<=1 and f2<=1
			assert abs(f1+f2-1.0) < 1e-6
			o.append(sd[s-1]*f1 + sd[s]*f2)
	return o



def test():
	assert Num.allclose(wp([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], None,
					[0.0, 1.0, 0.5, 0.51, 0.49, 0.01, 0.99]),
				[1.0, 10.0, 5.5, 5.6, 5.4, 1.0, 10.0], 0.0001)
	assert Num.allclose(wp([0, 1, 2, 3, 4], [0.1, 1.9, 1.9, 0.1, 1],
					[0.0, 1.0, 0.01, 0.02, 0.99]),
					[0.0, 4.0, 0.0, 0.05, 4.0], 0.0001)



if __name__ == '__main__':
	test()




"""This reads MLF (Master Label Files) for/from the HTK speech recognition
toolkit.
"""

import os.path
import glob
import die
from xwaves_errs import *

class ReferencedFileNotFound(Error):
	def __init__(self, *s):
		Error.__init__(self, *s)


def _findfile(f, postfix):
	if not f:
		return (None, None)
	d0, f0 = os.path.split(f)
	f0 = f0.rsplit('.', 1)[0]
	fx = os.path.join(d0, f0) + postfix
	gl = glob.glob(fx)
	if len(gl) == 0:
		raise ReferencedFileNotFound, fx
	assert len(gl) == 1, 'Too many alternatives for %s: %d' % (fx, len(gl))
	d1, f1 = os.path.split(gl[0])
	assert f1.endswith(postfix)
	f1 = f1[:-len(postfix)]
	return (d1, f1)


def _get_symbols(fd):
	sym = []
	while True:
		s = fd.readline()
		if s == '':
			# It is legal for the last chunk of a file
			# to end with a blank line, rather than a dot.
			if sym and sym[-1]=='':
				sym.pop()
				break
			else:
				raise BadFileFormatError
		s = s.rstrip('\r\n')
		if s == '.':
			break
		sym.append( s )
	return sym


def _get_threecol(fd):
	SCALE = 1e-7
	sym = []
	while True:
		s = fd.readline()
		if s == '':
			raise BadFileFormatError
		s = s.rstrip('\r\n')
		if s == '':
			# It is legal for the last chunk of a file
			# to end with a blank line rather than a
			# dot.
			if fd.readline() == '':
				break
			else:
				raise BadFileFormatError
		if s == '.':
			break
		t0, te, lab = s.split(None, 2)
		sym.append( (float(t0)*SCALE, float(te)*SCALE, lab) )
	return sym
	

def readiter(mlf_fn, postfix='.wav', datapath='.', strict=True,
		threecol=False, findfile=True, pathedit=None):
	"""Read a MLF file (relevant to HTK took kit).
	Returns [ {'filespec':path, 'd': d, 'f': f, 'symbols': zzz}, ... ]
	where 'd' and 'f' are only present if findfile==True.
	Datapath and pathedit are ways to deal with the
	situation where the MLF file has been moved, or (for other reasons)
	the filenames in the MLF file don't point to the actual data.

	"""
	if pathedit is None:
		pathedit = os.path.join

	dmlf, fmlf = os.path.split(mlf_fn)

	try:
		fd = open(mlf_fn, 'r')
	except IOError, x:
		raise NoSuchFileError(*(x.args))
	l = fd.readline()
	assert l=='#!MLF!#\n', 'l=%s' % l
	while True:
		f = fd.readline()
		f = f.strip()
		if f == '':
			break
		if f.startswith('"') and f.endswith('"'):
			f = f[1:-1]
		fspec = f
		rv = {'filespec': fspec}
		if findfile:
			try:
				d1, f1 = _findfile(pathedit(dmlf,datapath,f), postfix)
			except ReferencedFileNotFound, x:
				if strict:
					raise
				else:
					die.warn('No such file: %s from %s' % (x, fspec))
					_get_symbols(fd)
					continue
			else:
				rv['d'] = d1
				rv['f'] = f1
		if threecol:
			rv['symbols'] = _get_threecol(fd)
		else:
			rv['symbols'] = _get_symbols(fd)
		yield rv


def read(mlf_fn, **kw):
	return list( readiter(mlf_fn, **kw) )


class writer(object):
	SCALE = 1e7

	def __init__(self, mlf_fd):
		self.fd = mlf_fd
		self.fd.writelines('#!MLF!#\n')
		self.nchunks = 0

	def chunk(self, filespec, data):
		if self.nchunks > 0:
			self.fd.writelines('.\n')
		self.fd.writelines( [ '"%s"\n' % filespec,
					'\n'.join(data), '\n'
					]
				)
		self.nchunks += 1
		self.fd.flush()

	def threecol(self, filespec, tcdata):
		d = [ '%d %d %s' % (int(round(t0*self.SCALE)),
					int(round(te*self.SCALE)),
					lbl) 
			for (t0, te, lbl) in tcdata
			]
		self.chunk(filespec, d)

	def close(self):
		self.fd.writelines('\n')
		self.fd = None	# Normally, this will close the file.

	def __del__(self):
		if self.fd is not None:
			self.close()


if __name__ == '__main__':
	import sys
	THREECOL = False
	DATAPATH = None
	arglist = sys.argv[1:]
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '-threecol':
			THREECOL = True
		elif arg == '-datapath':
			DATAPATH = arglist.pop(0)
		else:
			die.die('Unrecognized argument: %s' % arg)
	for tmp in readiter(arglist[0], datapath=DATAPATH,
				threecol=THREECOL, findfile=(DATAPATH is not None)
				):
		print '[', tmp['filespec'], tmp.get('d', ''), tmp.get('f', ''), ']'
		for tmps in tmp['symbols']:
			if THREECOL:
				print tmps[0], tmps[1], tmps[2]
			else:
				print tmps
